import math
import numpy as np
from modelbase import ModelBase

class BearingsOnly(ModelBase):
    '''basically about being on a ww2 era sub. your sub is travelling along, and you begin hearing the sound of a ship at some bearing. over time, as the bearing changes, you can estimate the position and velocity of the ship. both your sub and the ship are moving the entire time. at some point you make a course change for your sub to pursue the ship.'''

    pass